from typing import List, Union

import numpy as np
import pandas as pd


def generate_synthetic_cohort_matrix(cohort_sizes: List[int], alphas: List[float], betas: List[float],
                                     max_obs_period: Union[int, float] = float('inf')) -> pd.DataFrame:
    cohorts = len(cohort_sizes)
    assert cohorts < max_obs_period

    cohort_matrix = []
    for cohort_number in range(cohorts):
        rand_data = np.random.geometric(
            np.random.beta(alphas[cohort_number],
                           betas[cohort_number],
                           size=cohort_sizes[cohort_number])
        )
        cohort = {
            i: 0
            for i in range(
                cohort_number + 1,
                min(max_obs_period, max(rand_data)) + 1
            )
        }
        for individual_death in rand_data:
            for alive_period in range(cohort_number + 1, min(max_obs_period, individual_death + cohort_number) + 1):
                cohort[alive_period] += 1
        cohort_matrix.append(cohort)

    return pd.DataFrame(cohort_matrix)




from typing import List, Any, Tuple

import theano
import theano.tensor as tt
import pymc3 as pm
import numpy as np
import pandas as pd
import arviz as az

from pymc3.theanof import floatX, intX
from scipy import stats
from scipy.special import beta as beta_f


def betaln(a, b):
    return tt.gammaln(a) + tt.gammaln(b) - tt.gammaln(a + b)


def beta_geom_llh(x, a, b):
    return betaln(a + 1, x + b - 1) - betaln(a, b)


def censored_beta_geom_llh(x, a, b):
    return betaln(a, b + x) - betaln(a, b)


def censored_beta_geom_pdf(x, a, b):
    return beta_f(a, b + x) / beta_f(a, b)


def beta_geom_pdf(x, a, b):
    return beta_f(a + 1, x + b - 1) / beta_f(a, b)


class SPShiftedBetaGeometric(stats.rv_continuous):
    def _pdf(self, x, t_max, a, b):
        if x >= t_max:
            return censored_beta_geom_pdf(x, a, b)
        return beta_geom_pdf(x, a, b)


class ShiftedBetaGeometric(pm.Discrete):
    """
    Pymc implementation of:
    Fader, Peter and Hardie, Bruce, How to Project Customer Retention (May 2006).
    Available at SSRN: https://ssrn.com/abstract=801145.
    or http://dx.doi.org/10.2139/ssrn.801145
    """

    def __init__(self, a, b, surviving, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.a = tt.as_tensor_variable(floatX(a))
        self.b = tt.as_tensor_variable(floatX(b))
        self.surviving = tt.as_tensor_variable(intX(surviving))
        self.mode = 1

    def logp(self, value):
        idx = tt.arange(1, value.shape[0] + 1, 1, dtype='int32')
        return tt.sum(value * beta_geom_llh(idx, self.a, self.b)) + self.surviving * censored_beta_geom_llh(
            value.shape[0], self.a, self.b)


def fit_sbg_model(cohort_matrix: pd.DataFrame, price: float = 1, all_users: List[int] = None,
                  min_cohort_size: int = 40, periods: int = 52, true_alpha=None, true_beta=None,
                  **sampler_kwargs: Any) -> Tuple[
    az.InferenceData, pm.Model]:
    """
    Fits the SBG model for customer LTV
    :param cohort_matrix: the retention matrix by cohort, see `generate_synthetic_cohort_matrix` for an example
    :param price: the price users pay for each subscription renewal, by default unitary price
    :param periods: the number of periods to forecast
    :param all_users: optional, adds a conversion rate component to the model if you have unconverted
        users using your product (e.g. "free users" or "free trials")
    :param sampler_kwargs: extra kwargs to be passed to the pymc sampler
    :param min_cohort_size: minimum size for a cohort to be considered in the fitting process
    :return: az.InferenceData of the sbg model
    """

    model = pm.Model()
    with model:
        alpha = pm.HalfCauchy('alpha', 0.05)  # pm.HalfFlat('alpha')#pm.HalfCauchy('beta', 1)
        beta = pm.HalfCauchy('beta', 0.05)  # pm.HalfFlat('beta')

        can_fit = False
        starting_payers = []
        for i in range(cohort_matrix.shape[0]):
            effective_data = cohort_matrix.values[i][~pd.isna(cohort_matrix.values[i])]
            starting_payers.append(effective_data[0])
            if len(effective_data) >= 4 and max(effective_data) > min_cohort_size:
                can_fit = True
                # Consider only cohorts with 4 FULL renewal periods and at least 40 (or min_cohort_size) ppl
                ShiftedBetaGeometric(
                    f'obs_{i}_cohort',
                    a=alpha,
                    b=beta,
                    observed=np.abs(np.diff(effective_data[:-1])),
                    surviving=effective_data[-2]
                )

        if not all_users:
            conversion_rate = pm.Deterministic('conversion_rate', tt.ones(1))
            pm.Deterministic('conversion_rate_by_cohort', tt.ones(cohort_matrix.shape[0]))

        else:
            if not all_users:
                all_users = starting_payers
            if not len(all_users) == cohort_matrix.shape[0]:
                raise Exception('If provided, free trials must be provided for each cohort')
            conversion_rate = pm.Uniform('conversion_rate', 0, 1)

            conversion_rate_alpha = pm.HalfCauchy('cohort_conversion_rate_alpha', 0.5)

            cohort_conversion_rate = pm.Beta('conversion_rate_by_cohort',
                                             alpha=conversion_rate_alpha,
                                             beta=conversion_rate_alpha * (1 / conversion_rate - 1),
                                             shape=len(all_users))

            pm.Binomial('conversion_rate_obs', n=all_users, p=cohort_conversion_rate, observed=starting_payers)

        survival_probabilities, _ = theano.scan(
            fn=lambda t: tt.switch(tt.gt(t, 1), sbg_survival_rate(t - 1, alpha, beta), tt.ones(1)),
            sequences=tt.arange(start=1, stop=periods + 1, step=1),
        )

        survival_curve = tt.cumprod(survival_probabilities)
        pm.Deterministic('survival_curve', survival_curve)
        pm.Deterministic('ltv', tt.cumsum(survival_curve * conversion_rate * price))

        if true_alpha and true_beta:
            survival_probabilities, _ = theano.scan(
                fn=lambda t: tt.switch(tt.gt(t, 1), sbg_survival_rate(t - 1, true_alpha, true_beta), tt.ones(1)),
                sequences=tt.arange(start=1, stop=periods + 1, step=1),
            )

            survival_curve = tt.cumprod(survival_probabilities)
            pm.Deterministic('true_survival_curve', survival_curve)
            pm.Deterministic('true_ltv', tt.cumsum(survival_curve * conversion_rate * price))

        if not can_fit:
            raise Exception(
                f'Not enough data for fitting, you need 1 cohort with 4 observation periods and >{min_cohort_size}ppl')

        data = pm.sample(return_inferencedata=True, **sampler_kwargs)

        samples = pm.fast_sample_posterior_predictive(data.posterior)

    return az.concat(data, az.from_pymc3(posterior_predictive=samples)), model


def sbg_survival_rate(x, a, b):
    """
    Probability that a customer alive at time x-1 is still alive at time x
    """
    return (b + x - 1) / (a + b + x - 1)


def _weighted_cohort_line(cohort_matrix):
    z = {}
    for idx, cohort_line in enumerate(cohort_matrix.values):
        for cohort_period, population in enumerate(cohort_line[idx:]):
            if not pd.isna(population):
                z.setdefault(cohort_period, [])
                z[cohort_period].append((population / cohort_line[~pd.isna(cohort_line)].max(), population))
    out = []
    for k, v in z.items():
        tw = 0
        tv = []
        for pct, weight in v:
            tv.append(pct * weight)
            tw += weight
        out.append(sum(tv) / tw)
    return np.array(out)


def compute_empirical_ltv(cohort_matrix: pd.DataFrame, price: float = 1, conversion_rate: float = 1,
                          periods: int = 52) -> Tuple[np.array, np.array, np.array]:
    """
    :param conversion_rate: optional, if the model was fit using it.
    :param cohort_matrix: the retention matrix by cohort, see `generate_synthetic_cohort_matrix` for an example
    :param price: the price your users pays for each renewal, remember to consider taxes
    :param periods: the maximum number of computed periods returned by the function
    :return: Returns three np.arrays containing lower, median and upper credible intervals for the empirical LTV
    """
    weighted_cohort = _weighted_cohort_line(cohort_matrix)
    empirical_ltv = np.cumsum(weighted_cohort * price) * conversion_rate
    return empirical_ltv[:periods]



import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import pandas as pd
import numpy as np
import arviz as az


def plot_ltv(empirical_ltv, inference_data=None, hdi_prob=.95, extra_label_text='', ax=None):
    if ax is None:
        fig, ax = plt.subplots(figsize=(20, 10))
    if inference_data:
        az.plot_hdi(x=np.arange(52), hdi_prob=hdi_prob, smooth=False, y=inference_data['posterior']['ltv'], ax=ax)
        curve_m = np.median(np.median(inference_data['posterior']['ltv'], axis=0), axis=0)
        ax.plot(curve_m, 'k', linestyle='dashed', alpha=0.5,
                label=f'Median ltv: {curve_m[len(curve_m) - 1].round(2)}')

    if 'true_ltv' in inference_data['posterior'].keys():
        curve_m = np.median(np.median(inference_data['posterior']['true_ltv'], axis=0), axis=0)
        ax.plot(curve_m, 'k', alpha=0.5,
                label=f'True ltv: {curve_m[len(curve_m) - 1].round(2)}')

    ax.plot(empirical_ltv, 'o',
            label=f'{extra_label_text}Empirical @{len(empirical_ltv)} periods: {empirical_ltv[len(empirical_ltv) - 1].round(2)}')
    return ax


def plot_conversion_rate(inference_data, hdi_prob=.95, extra_label_text='', ax=None):
    if ax is None:
        fig, ax = plt.subplots(figsize=(20, 10))

    conversion_rate_by_cohort = inference_data['posterior']['conversion_rate_by_cohort']
    az.plot_hdi(x=np.arange(conversion_rate_by_cohort.shape[-1]), hdi_prob=hdi_prob, smooth=False,
                y=conversion_rate_by_cohort, ax=ax)
    curve_m = np.median(np.median(inference_data['posterior']['conversion_rate_by_cohort'], axis=0), axis=0)
    ax.plot(curve_m, 'k-', alpha=0.3,
            label=f'{extra_label_text}Median Conversion Rate: {np.median(curve_m).round(2)}')
    ax.plot(curve_m, 'ko', alpha=0.8,
            label=f'{extra_label_text}Median Conversion Rate: {np.median(curve_m).round(2)}')


def plot_cohort_matrix_retention(cohort_matrix, title=''):
    cohort_size = cohort_matrix.max(axis=1)
    retention_matrix = cohort_matrix.divide(cohort_size, axis=0)

    with sns.axes_style("white"):
        fig, ax = plt.subplots(1, 2, figsize=(20, 10), sharey='all', gridspec_kw={'width_ratios': [1, 11]})
        sns.heatmap(retention_matrix.iloc[:, :-1],
                    mask=retention_matrix.iloc[:, :-1].isnull(),
                    annot=True,
                    fmt='.0%',
                    cmap='RdYlGn',
                    ax=ax[1])
        ax[1].set_title(title, fontsize=12)
        ax[1].set(xlabel='# of periods',
                  ylabel='')

        # cohort size
        cohort_size_df = pd.DataFrame(cohort_size).rename(columns={0: 'cohort_size'})
        white_cmap = mcolors.ListedColormap(['white'])
        sns.heatmap(cohort_size_df,
                    annot=True,
                    cbar=False,
                    fmt='g',
                    cmap=white_cmap,
                    ax=ax[0])

        fig.tight_layout()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lucius_ltv']

package_data = \
{'': ['*']}

install_requires = \
['arviz>=0.11.4,<0.12.0', 'pymc3>=3.11.4,<4.0.0', 'seaborn>=0.11.2,<0.12.0']

setup_kwargs = {
    'name': 'lucius-ltv',
    'version': '0.1.1',
    'description': 'A simple multicohort LTV calculator for subscription based products',
    'long_description': '# Lucius LTV\n\nA Python simple multi-cohort LTV calculator library for subscription-based products.\n\n## Installation\n\n```shell\npip install lucius-ltv\n```\n\n## Building a cohort matrix\n\nTo use this library, the first step is to build a **Cohort retention matrix** for your product.\n\nTo build a cohort matrix:\n * For each period of interest (week/month/year - depending on your subscription)\n * Retrieve the number of users that started paying for the first time\n * Track those users over time to observe how many where still alive in the following periods\n * Stack them in order to obtain a pd.DataFrame resembling this:\n\n| 2015 | 2016 | 2017 | 2018 | 2019 | 2020 | 2021 | 2022 |\n|------|------|------|------|------|------|------|------|\n| 100  | 95   | 82   | 78   | 75   | 71   | 68   | 63   |\n|      | 120  | 109  | 99   | 87   | 80   | 67   | 66   |\n|      |      | 101  | 90   | 80   | 77   | 73   | 68   |\n|      |      |      | 130  | 122  | 115  | 108  | 99   |\n|      |      |      |      | 95   | 91   | 85   | 67   |\n|      |      |      |      |      | 102  | 90   | 81   |\n|      |      |      |      |      |      | 90   | 80   |\n\nAlternatively if you are just trying out this library, you can generate a retention cohort matrix by calling `generate_synthetic_cohort_matrix`. An example of this is contained in the sample file, `sample.py` shipped with this project.\n\n## Fitting the model\n\nOnce you have a cohort matrix, you can finally fit the sBG model that will predict LTV. \nBehind the scenes this library implements "Fader, Peter and Hardie, Bruce, How to Project Customer Retention (May 2006)" (Available at SSRN: https://ssrn.com/abstract=801145\nor http://dx.doi.org/10.2139/ssrn.801145) using the powerful Bayesian Inference library [pymc3](https://docs.pymc.io/en/v3/).\n\nTo fit the pymc model, simply run:\n\n### Basic fit\n```python\ninference_data, model = fit_sbg_model(\n    cohort_matrix,\n    periods=10,    # Number of projected periods\n)\n```\n\n### Fit with "all_users"\n\nIf your product offers free trials or initial offers of any kind, you can add a conversion layer to the model, by specifying the starting users:\n```python\ninference_data, model = fit_sbg_model(\n    cohort_matrix,\n    all_users=[150, 147, 180, 160, 130, 140, 160], \n    periods=10,    # Number of projected periods\n)\n```\n\n### Fit with true parameters\n\nIf you are testing the model with synthetic data you can specify the true alpha/beta sbg parameter values (those won\'t be used in fitting, but to generate reference/comparison timeseries)\n\n```python\ninference_data, model = fit_sbg_model(\n    cohort_matrix,\n    periods=10,    # Number of projected periods\n    true_a=1.5,\n    true_b=2.3,\n)\n```\n\n### Extra pymc sampling parameters\n\nIf you want you can pass extra arguments that will be re-routed to the pymc sampler, like `target_accept`, `steps`, `tune`, etc..\n\n## Analyzing results\n\nThe returned `inference_data` is a standard Arviz InferenceData object will contain a posterior estimate of the modelled variables as returned by pymc3.\nThe most interesting variables are:\n\n* `inference_data.posterior[\'ltv\']` the LTV timeseries\n* `inference_data.posterior[\'conversion_rate_by_cohort\']` the conversion rates if you did the fit with `all_users`\n* `inference_data.posterior[\'ltv\']` the true LTV timeseries if you provided true values of alpha and beta\n\nPlease note that these are trace objects resulting from the sampling process, \ntherefore each variable will have multiple possible values representing the sampling of the posterior distribution. \nFrom this sampling you can obtain high-density intervals using [`np.percentile`](https://numpy.org/doc/stable/reference/generated/numpy.percentile.html) or [`az.hdi`](https://arviz-devs.github.io/arviz/api/generated/arviz.hdi.html).\n\n### Empirical LTV from cohort matrix\n\nYou can also compute the empirical LTV starting from the cohort matrix, allowing you to compare empirical LTV to the projected one.\n\n```python\nempirical_ltv = compute_empirical_ltv(cohort_matrix=cohort_matrix)\n```\n\n## Plots\n\nThe library also includes a few methods for quickly plotting results.\n\n### LTV\n\nPlot the user lifetime value\n\n```python\nfig, ax = plt.subplots(figsize=(20, 10))\nplot_ltv(empirical_ltv, inference_data=inference_data, ax=ax)\nfig.suptitle(f\'Lifetime value\')\nax.legend()\nax.grid()\n```\n\n### Conversion Rate\n\nPlot the conversion rate by cohort\n\n```python\nfig, ax = plt.subplots(figsize=(20, 10))\nplot_conversion_rate(inference_data, ax=ax)\nfig.suptitle(f\'Conversion Rate\')\nax.grid()\n```\n\n### Cohort matrix\n\nPlot the cohort matrix retention rates\n\n```python\nplot_cohort_matrix_retention(cohort_matrix, \'Cohort Retention\')\n```\n\n### Posterior distributions vs true values\n\nPlot the posterior distributions vs true values \n\n```python\nfig, (ax1, ax2, ax3) = plt.subplots(3, figsize=(12, 12))\naz.plot_posterior(inference_data, var_names=("alpha",),\n                  ref_val=true_alpha,\n                  ax=ax1)\naz.plot_posterior(inference_data, var_names=("beta",),\n                  ref_val=true_beta,\n                  ax=ax2)\naz.plot_posterior(inference_data, var_names=("conversion_rate",),\n                  ref_val=1/true_conversion_rate,\n                  ax=ax3)\nfig.suptitle(f\'True v Recovered values\')\n```',
    'author': 'Luca Giacomel',
    'author_email': 'luca.giacomel@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/luke14free/lucius-ltv',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)
